#include "anisodamrot.h"
#include "matrix.h"
#include "vector.h"
#include "elastisomat.h"
#include "global.h"
#include "intpoints.h"
#include "tensor.h"
#include "vecttens.h"
#include <math.h>
#include <stdlib.h>

#define nijac 20
#define limit 1.0e-16
#ifndef M_PI
 #define M_PI 3.14159265358979323846
#endif





/**
  This constructor inializes attributes to zero values.
  
  TKo 9.2006
*/
anisodamrot::anisodamrot (void)
{
  cde = corr_off;
  ac = 0.0;  bc = 0.0;  y0c = 0.0;  at = 0.0;  bt = 0.0;  y0t = 0.0;
  a = 0.0;  b = 0.0;  y0 = 0.0; gfc = 0.0; gft = 0.0; gf = 0.0;
}

/**
  This destructor is only for the formal purposes.
  
  TKo 9.2006
*/
anisodamrot::~anisodamrot (void)
{

}

/**
  This function reads material parameters from the opened text file given
  by the parameter in.

  @param in - pointer to the opened input text file
  
  TKo 9.2006
*/
void anisodamrot::read (XFILE *in)
{
  xfscanf (in, "%d", (int *)&cde);
  switch (cde)
  {
    case corr_off:
      xfscanf (in, "%lf %lf %lf", &ac, &bc, &y0c);
      xfscanf (in, "%lf %lf %lf", &at, &bt, &y0t);
      xfscanf (in, "%lf %lf %lf", &a, &b, &y0);
      break;
    case corr_on:
      xfscanf (in, "%lf %lf %lf", &gfc , &bc, &y0c);
      xfscanf (in, "%lf %lf %lf", &gft, &bt, &y0t);
      xfscanf (in, "%lf %lf %lf", &gf, &b, &y0);
      break;
    default:
      fprintf(stderr, "\nUnknown type of correction of disipated energy is required\n");
      fprintf(stderr, "in function anisodamrot::read (file %s, line : %d)\n", __FILE__, __LINE__);
  }
}



/** 
    Function computes dimensionless damage driving force for volumetric damage parameter.

    @param  nu   - Poissons ratio
    @param  peps - %vector of principal strains
    
    @retval Function returns real value of damage driving force.

    TKo - 9.2006
*/
double anisodamrot::damdrvforce_vol(double nu, vector &peps)
{
  long i;
  double ev = 0.0;
  double k0_prime, g0_prime, ret;  
  // strain tensor due to positive principal strains
  for (i = 0; i < 3; i++)
    ev += peps(i);
  ev /= 3.0;

  // take only positive volumetric strain  
  if (ev <= 0.0)
    return 0.0;

  // positive dimensionless volumetric damage driving force
  k0_prime = 1.0/(1.0-2.0*nu);
  g0_prime = 1.0/(2.0*(1.0+nu));
  ret = 3.0/2.0*(k0_prime-2.0*g0_prime)*ev*ev;  
  return ret;
}

/** 
    Function computes dimensionless damage driving force for deviatoric damage parameters.

    @param  nu   - Poissons ratio
    @param  peps - %vector of principal strains
    @param  t    - %matrix of principal directions stored in colomns for given
                   principal strains
    @param  pyt  - %vector of principal values of driving forces for tension damage parameters
    @param  pyc  - %vector of principal values of driving forces for compression damage parameters        

    TKo - 9.2006
*/
void anisodamrot::damdrvforce_dev(double nu, vector &peps, matrix &t, vector &pyt, vector &pyc, matrix &eps_t, matrix &eps_c)
{
  long i;
  matrix tyt(3,3), tyc(3,3);  
  double g0_prime;
  g0_prime = 1.0/(2.0*(1.0+nu));  

  // strain tensor due to positive principal strains
  fillm(0.0, eps_t);
  fillv(0.0, pyt);
  for (i = 0; i < 3; i++)
  {
    if (peps[i] > 0.0)
    {
      eps_t[i][i] = peps[i];
      pyt[i] = g0_prime*peps[i]*peps[i];
    }
  }
  lgmatrixtransf(eps_t, t);
  // strain tensor due to negative principal strains
  fillm(0.0, eps_c);
  for (i = 0; i < 3; i++)
  {
    if (peps[i] < 0.0)
    {
      eps_c[i][i] = peps[i];
      pyc[i] = g0_prime*peps[i]*peps[i];
    }
  }
  lgmatrixtransf(eps_c, t);
}

/**
  Function computes value of load function for the volumetric damage.

  @param  y  - volumteric dimensionless damage driving force
  @param  d  - actual value of damage parameter
  @param  aa - actual value of material parameter a

  @retval The function returns value of load function for the volumetric damage.
  
  TKo 9.2006
*/
double anisodamrot::loadfuncvol(double y, double d, double aa)
{
  double f;
  double tmp1, tmp2;

  if (y <= y0) 
   f = -1.0;      
  else
  {
    // compute (y-y0)^b
    tmp1 = y - y0;
    tmp2 = log(tmp1);
    tmp1 = exp(b*tmp2);
    // compute value of load function
    f = (1.0 - d)*(1.0+aa*tmp1) - 1.0;
  }
  return f;
}


/**
  Function computes value of load function for deviatoric damage

  @param  pyt  - %vector of principal dimensionless damage driving force for tension
  @param  pyc  - %vector of principal dimensionless damage driving force for tension
  @param  damt - %vector actual values of principal damage parameters for tension
  @param  damc - %vector actual values of principal damage parameters for compression
  @param  aat  - actual values of material parameter a for tension 
  @param  aac  - actual values of material parameter a for compression 
  @param  lft  - %vector of resulting values of load function for tension for each principal direction
  @param  lfc  - %vector of resulting values of load function for compression for each principal direction

  @retval The function returns load function values separately for tension and compression
          via parameters lft and lfc for each principal direction.

  TKo 9.2006
*/
void anisodamrot::loadfuncdev(vector &pyt, vector &pyc, vector &damt, vector &damc, 
                           double aat, double aac, vector &lft, vector &lfc)
{
  long i;
  double tmp1, tmp2;

  for (i=0; i<3; i++)
  {
    // for tension
    if (pyt[i] <= y0t) 
      lft[i] = -1.0;      
    else
    {
      // compute (yt-y0t)^bt
      tmp1 = pyt[i] - y0t;
      tmp2 = log(tmp1);
      tmp1 = exp(bt*tmp2);
      // compute value of load function for tension in given principal direction i
      lft[i] = (1.0 - damt[i])*(1.0+aat*tmp1) - 1.0;
    }
    // for compression
    if (pyc[i] <= y0c) 
      lfc[i] = -1.0;      
    else
    {
      // compute (yc-y0c)^bc
      tmp1 = pyc[i] - y0c;
      tmp2 = log(tmp1);
      tmp1 = exp(bc*tmp2);
      // compute value of load function for compression in given principal direction i
      lfc[i] = (1.0 - damc[i])*(1.0+aac*tmp1) - 1.0;
    }
  }  
  return;
}


/**
  Function computes value of damage parameter d. The damage parameter
  is derived from the damage function.

  @param y   - driving forces for volumetric damage
  @param aa  - actual value of material parameter a 

  @retval The function returns value of the volumetric damage parameter.
  
  TKo 9.2006
*/
double anisodamrot::dam_vol(double y, double aa)
{
  double tmp, dam = 0.0;

  if (y > y0)
  {  
    tmp = log(y-y0);
    dam = aa*exp(tmp*b);
    tmp = 1.0/(1.0+dam);
    dam = 1.0-tmp;
  }
  return dam;
}


/**
  Function computes values of principal damage parameters Dt and Dc. The damage parameters
  are derived from the damage function.

  @param ipp - integration point number
  @param pyc - %vector of principal driving forces for tension
  @param pyt - %vector of principal driving forces for tension
  @param aat - actual value of material parameter at 
  @param aac - actual value of material parameter ac 
  @param pdamc - %vector of principal damage parameters for tension
  @param pdamt - %vector of principal damage parameters for compression
  
  TKo 9.2006
*/
void anisodamrot::pdam_dev (vector &pyc, vector &pyt, double aat, double aac, 
                             vector &pdamt, vector &pdamc)
{
  long i;
  double tmp, ltmp;

  fillv(0.0, pdamt);
  fillv(0.0, pdamc);

  for(i=0; i<3; i++)
  {    
    // tension 
    if (pyt[i] > y0t)
    {
      ltmp = log(pyt[i] - y0t);
      pdamt[i] = aat*exp(ltmp*bt);
      tmp = 1.0/(1.0+pdamt[i]);
      pdamt[i] = 1.0 - tmp;
    }
    // compression 
    if (pyc[i] > y0c)
    {
      ltmp = log(pyc[i] - y0c);
      pdamc[i] = aac*exp(ltmp*bc);
      tmp = 1.0/(1.0+pdamc[i]);
      pdamc[i] = 1.0-tmp;
    }
  }
}



/**
  Function computes material parameters A, At and Ac. 
  If the correction of dissipated energy is required, they are computed with respect 
  to size of element for given ipp and variable softening modulus technique was used in 
  order to avoid mesh dependence. Otherwise, values read from the input file are used.

  @param d - allocated matrix structure for material stiffness %matrix
  @param ipp - integration point number
  @param ido - index of internal variables for given material in the ipp other array
  @param aa -  actual value of parameter A 
  @param aat - actual value of parameter Ac 
  @param aac - actual value of parameter Ac 

  @retval The function returns computed 
  
  TKo 9.2006
*/
void anisodamrot::give_actual_param_a(long ipp, long ido, double &aa, double &aat, double &aac)
{
  long eid;
  double e, h, tmp;

  aa  = Mm->ip[ipp].eqother[ido];
  aat = Mm->ip[ipp].eqother[ido+1];
  aac = Mm->ip[ipp].eqother[ido+2];
  if ((aa == 0.0) || (aac == 0.0) || (aat == 0.0))
  {
    if (cde == corr_off)
    // no correction of dissipated energy is required
    {
      aa  = a;
      aat = at;
      aac = ac;
      Mm->ip[ipp].eqother[ido]   = aa;
      Mm->ip[ipp].eqother[ido+1] = aat;
      Mm->ip[ipp].eqother[ido+2] = aac;
      return;
    }
    eid = Mm->elip[ipp];
    switch (Mt->give_dimension(eid))
    {
      case 1 :
        h = Mt->give_length(eid);
        break;
      case 2 :
        h = sqrt(Mt->give_area(eid));
        break;
      case 3 :
        h = pow(Mt->give_volume(eid), 1.0/3.0);
        break;
      default :
        fprintf(stderr, "\nError determining dimension of element");
        fprintf(stderr, " in function anisodamrot::give_actual_param_a\n(file %s, line %d, elem : %ld)\n", __FILE__, __LINE__, eid+1);
    }
    e = Mm->give_actual_ym(ipp);
    // variable softening for volumetric damage
    tmp = (gf/h - y0/e)*b*e*sin(M_PI/b)/M_PI; 
    if (tmp < 0.0)
    {
      fprintf(stderr, "\nError while computing variable softening for volumetric damage");
      fprintf(stderr, " in function anisodamrot::give_actual_param_a\n(file %s, line %d, elem : %ld)\n", __FILE__, __LINE__, eid+1);     
    }
    aa = pow(tmp, -b);
    // variable softening for tension
    tmp = (gft/h - y0t/e)*bt*e*sin(M_PI/bt)/M_PI; 
    if (tmp < 0.0)
    {
      fprintf(stderr, "\nError while computing variable softening for tension");
      fprintf(stderr, " in function anisodamrot::give_actual_param_a\n(file %s, line %d, elem : %ld)\n", __FILE__, __LINE__, eid+1);     
    }
    aat = pow(tmp, -bt);
    // variable softening for compression
    tmp = (gfc/h - y0c/e)*bc*e*sin(M_PI/bc)/M_PI; 
    if (tmp < 0.0)
    {
      fprintf(stderr, "\nError while computing variable softening for compression");
      fprintf(stderr, " in function anisodamrot::give_actual_param_a\n(file %s, line : %d, elem : %ld)\n", __FILE__, __LINE__, eid+1);     
    }
    aac = pow(tmp, -bc);
    Mm->ip[ipp].eqother[ido]   = aa;
    Mm->ip[ipp].eqother[ido+1] = aat;
    Mm->ip[ipp].eqother[ido+2] = aac;
  }
}



/**
  Function computes initializes material parameters A, At and Ac. 

  @param d - allocated matrix structure for material stiffness %matrix
  @param ipp - integration point number
  @param ido - index of internal variables for given material in the ipp other array

  @retval The function returns computed 
  
  TKo 9.2006
*/
void anisodamrot::initvalues(long ipp, long ido)
{
  double aa, aat, aac;
  give_actual_param_a(ipp, ido, aa, aac, aat);
}



/**
  This function computes material stiffnes %matrix.

  @param d - allocated matrix structure for material stiffness %matrix
  @param ipp - integration point number
  @param ido - index of internal variables for given material in the ipp other array
  
  TKo
*/
void anisodamrot::matstiff (matrix &d,long ipp,long ido)
{
//  long ncompstr = Mm->ip[ipp].ncompstr;

  switch (Mp->stmat)
  {
    case initial_stiff :
      elmatstiff(d, ipp);
      break;
    case tangent_stiff :
      elmatstiff(d, ipp);
/*      dp=Mm->ip[ipp].eqother[ido+2*ncompstr+2];
      if (dp > 0.999999)
        dp = 0.999999;
      cmulm (1.0-dp,d);*/
      break;
    default :
      fprintf(stderr, "\n\nError - unknown type of stifness matrix");
      fprintf(stderr, "\n in function anisodamrot::matstiff (file %s, line %d)\n", __FILE__, __LINE__);
  }
}


/**
  This function computes elastic material stiffnes %matrix from actual Young modulus.

  @param d - allocated matrix structure for material stiffness %matrix
  @param ipp - integration point number

  TKo
*/
void anisodamrot::elmatstiff (matrix &d,long ipp)
{
  double e, e0;
  long idem = Mm->ip[ipp].gemid();

  Mm->elmatstiff (d,ipp);
  e = Mm->give_actual_ym(ipp);
  if (Mm->ip[ipp].tm[idem] != elisomat)
  {
    fprintf(stderr, "\n\n Invalid type of elastic material is required");
    fprintf(stderr, "\n  in function anisodamrot::elmatstiff, (file %s, line %d)\n", __FILE__, __LINE__);
  }
  e0 = Mm->eliso[Mm->ip[ipp].idm[idem]].e;
  cmulm(e/e0, d);
}




/**
  This function computes correct stresses in the integration point and stores
  them into ip stress array.

  @param ipp - integration point pointer
  @param im  - index of material type for given ip
  @param ido - index of internal variables for given material in the ipp other array

  TKo, 9.2006
*/
void anisodamrot::nlstresses (long ipp, long im, long ido)
{
  long ime, idem, i;
  long ncompstr=Mm->ip[ipp].ncompstr;

  // actual material parameters a 
  double aa, aat, aac;
  // old values of damage parameters
  double dvo;
  matrix dto(3,3), dco(3,3), dold(3,3);
  // decomposed strains to tensional and compressive parts
  matrix eps_t(3,3), eps_c(3,3);
  // new values of driving forces
  double yn;
  vector pyt(3), pyc(3);
  // damage parameters
  double dv;
  vector pdt(3), pdc(3);
  vector pdto(3), pdco(3);
  // strains and principal strains
  vector epsn(ncompstr);
  // tensor form of strains and transformation matrix for principal direction of strains
  matrix epst(3,3), t(3,3), tomega(3,3);
  // stress tensor and stress vector
  matrix sigt(3,3);
  vector sigma(ncompstr);
  // bulk modulus*3, shear modulus, volumetric strain, Young modulus, Poissons ratio
  double k0, g0, epsv, e, nu;
  // actual values of load fuctions
  vector lft(3), lfc(3);
  // principal strains
  vector peps(3);
  // damage tensors
  matrix dt(3,3), dc(3,3), dn(3,3);
  // auxiliary matrices
  matrix tmp1(3,3), tmp2(3,3);
  matrix omegat(3,3), omegac(3,3);

  e = Mm->give_actual_ym(ipp);
  idem = Mm->ip[ipp].gemid();
  nu = Mm->eliso[Mm->ip[ipp].idm[idem]].nu;
  if (Mm->ip[ipp].ssst == planestress)
    Mm->ip[ipp].strain[3] = -nu / (1.0 - nu) * (Mm->ip[ipp].strain[0]+Mm->ip[ipp].strain[1]);
  // initialize values of material parameter a
  give_actual_param_a(ipp, ido, aa, aat, aac);

  //  total new strains
  for (i=0;i<ncompstr;i++)
    epsn[i] = Mm->ip[ipp].strain[i];

  // previous damage parameters
  dvo = Mm->ip[ipp].eqother[ido+3]; // volumetric damage parameter
  // deviatoric damage parameters for tension
  dto[0][0] = Mm->ip[ipp].eqother[ido+3+1+0];
  dto[0][1] = dto[1][0] = Mm->ip[ipp].eqother[ido+3+1+1];
  dto[0][2] = dto[2][0] = Mm->ip[ipp].eqother[ido+3+1+2];
  dto[1][1] = Mm->ip[ipp].eqother[ido+3+1+3];
  dto[1][2] = dto[2][1] = Mm->ip[ipp].eqother[ido+3+1+4];
  dto[2][2] = Mm->ip[ipp].eqother[ido+3+1+5];
  // deviatoric damage parameters for compression
  dco[0][0] = Mm->ip[ipp].eqother[ido+3+1+6+0];
  dco[0][1] = dco[1][0] = Mm->ip[ipp].eqother[ido+3+1+6+1];
  dco[0][2] = dco[2][0] = Mm->ip[ipp].eqother[ido+3+1+6+2];
  dco[1][1] = Mm->ip[ipp].eqother[ido+3+1+6+3];
  dco[1][2] = dco[2][1] = Mm->ip[ipp].eqother[ido+3+1+6+4];
  dco[2][2] = Mm->ip[ipp].eqother[ido+3+1+6+5];
  
  // new driving forces
  vector_tensor(epsn, epst, Mm->ip[ipp].ssst, strain);
  princ_val (epst,peps,t,nijac,limit,Mp->zero,3,1);
  yn = damdrvforce_vol(nu, peps);
  damdrvforce_dev(nu, peps, t, pyt, pyc, eps_t, eps_c);

  // new damage parameters
  dv = dam_vol(yn, aa);
  pdam_dev(pyc, pyt, aat, aac, pdt, pdc);

  // transformation of principal damage to the global coordinate system
  fillm(0.0, dt);
  fillm(0.0, dc);
  glmatrixtransf(dto, t);
  glmatrixtransf(dco, t);
  copym(dco, dc);
  copym(dto, dt);
  for (i = 0; i < 3; i++)
  {
    if (dc[i][i] < pdc[i])
      dc[i][i] = pdc[i];
    if (dt[i][i] < pdt[i])
      dt[i][i] = pdt[i];
  }
  lgmatrixtransf(dt, t);
  lgmatrixtransf(dc, t);

  // damage parameters can be increased only
  if (dv < dvo)
    dv = dvo;
    
  // checking of load functions for deviatoric damage tensors
  // after previous conditons (enabling rotaion of principal directions)
  long err;
  copym(dt, tmp1);
  copym(dc, tmp2);
  glmatrixtransf(tmp1, t);
  glmatrixtransf(tmp2, t);
  for (i = 0; i < 3; i++)
  {
    pdt[i] = tmp1[i][i];
    pdc[i] = tmp2[i][i];
  }
  loadfuncdev(pyt, pyc, pdt, pdc, aat, aac, lft, lfc);
  err = 0;
  for (i =0; i<3; i++)
  {
    if (lft[i] > 1.0e-3)
      err = 1;
    if (lfc[i] > 1.0e-3)
      err = 2;
  }
  if (err)
  {
    print_err("Load function in anisotropic damage is not in tolerance\n", __FILE__, __LINE__, __func__);
    abort();
  }


  // compute Omega_t=(1-Dt)^(1/2) and Omega_c=(1-Dc)^(1/2)
  copym(dt, omegat);
  cmulm(-1.0, omegat);
  for(i=0; i<3; i++)
    omegat[i][i] += 1.0;
  princ_val (omegat,pdt,tomega,nijac,limit,Mp->zero,3,1);
  fillm(0.0, omegat);
  for(i=0; i<3; i++)
  {
    if (pdt[i] < 0.0)
    {
      print_err("Principal value of damage tensor in tension > 1.0", __FILE__, __LINE__, __func__);
      pdt[i] = 0.0;
    }
    omegat[i][i] = sqrt(pdt[i]);
  }
  lgmatrixtransf(omegat, tomega);

  copym(dc, omegac);
  cmulm(-1.0, omegac);
  for(i=0; i<3; i++)
    omegac[i][i] += 1.0;
  princ_val (omegac,pdc,tomega,nijac,limit,Mp->zero,3,1);
  fillm(0.0, omegac);
  for(i=0; i<3; i++)
  {
    if (pdc[i] < 0.0)
    {
      print_err("Principal value of damage tensor in tension > 1.0", __FILE__, __LINE__, __func__);
      pdc[i] = 0.0;
    }
    omegac[i][i] = sqrt(pdc[i]);
  }
  lgmatrixtransf(omegac, tomega);
      
  // new stresses
  e  = Mm->give_actual_ym(ipp);
  idem = Mm->ip[ipp].gemid();
  ime = Mm->ip[ipp].idm[idem];
  nu = Mm->eliso[ime].nu;
  k0 = e;
  k0 /= (1-2.0*nu);
  g0 = e;
  g0 /= 2.0*(1.0+nu);
  // compute volumetric part : 3/2*(ko-2*Go)*(1-d)*eps_v
  epsv = peps[0]+peps[1]+peps[2];
  epsv /= 3.0;
  sigt[0][0] = k0 - 2.0*g0;
  if (epsv > 0.0)
    epsv -= dv*epsv;
  sigt[0][0] *= epsv;
  sigt[1][1] = sigt[2][2] = sigt[0][0];
  // compute deviatoric part for tension : 2*Go*Omega_t*eps_t*Omega_t
  mxm(omegat, eps_t, tmp1);
  mxm(tmp1, omegat, tmp2);
  cmulm(2.0*g0, tmp2);
  addm(sigt, tmp2, sigt);
  // compute deviatoric part for compression : 2*Go*Omega_c*eps_c*Omega_c
  mxm(omegac, eps_c, tmp1);
  mxm(tmp1, omegac, tmp2);
  cmulm(2.0*g0, tmp2);
  addm(sigt, tmp2, sigt);
  tensor_vector(sigma, sigt, Mm->ip[ipp].ssst, stress);
  for (i=0;i<ncompstr;i++)
    Mm->ip[ipp].stress[i]=sigma[i];

  //  new data storage
  Mm->ip[ipp].other[ido+3] = dv; // volumetric damage parameter
  // deviatoric damage parameters for tension
  Mm->ip[ipp].other[ido+3+1+0] = dt[0][0];
  Mm->ip[ipp].other[ido+3+1+1] = dt[0][1];
  Mm->ip[ipp].other[ido+3+1+2] = dt[0][2];
  Mm->ip[ipp].other[ido+3+1+3] = dt[1][1];
  Mm->ip[ipp].other[ido+3+1+4] = dt[1][2];
  Mm->ip[ipp].other[ido+3+1+5] = dt[2][2];
  // deviatoric damage parameters for compression
  Mm->ip[ipp].other[ido+3+1+6+0] = dc[0][0];
  Mm->ip[ipp].other[ido+3+1+6+1] = dc[0][1];
  Mm->ip[ipp].other[ido+3+1+6+2] = dc[0][2];
  Mm->ip[ipp].other[ido+3+1+6+3] = dc[1][1];
  Mm->ip[ipp].other[ido+3+1+6+4] = dc[1][2];
  Mm->ip[ipp].other[ido+3+1+6+5] = dc[2][2];
  // principal values of damage tensor for tension
  copym(dt, omegat);
  princ_val (omegat,pdt,tomega,nijac,limit,Mp->zero,3,1);
  Mm->ip[ipp].other[ido+3+1+6+6+0] = pdt[0];
  Mm->ip[ipp].other[ido+3+1+6+6+1] = pdt[1];
  Mm->ip[ipp].other[ido+3+1+6+6+2] = pdt[2];
  // principal values of damage tensor for compression
  copym(dc, omegac);
  princ_val (omegac,pdc,tomega,nijac,limit,Mp->zero,3,1);
  Mm->ip[ipp].other[ido+3+1+6+6+3+0] = pdc[0];
  Mm->ip[ipp].other[ido+3+1+6+6+3+1] = pdc[1];
  Mm->ip[ipp].other[ido+3+1+6+6+3+2] = pdc[2];
}

/**
  Function updates values in the other array reached in the previous equlibrium state to
  values reached in the new actual equilibrium state.

  @param ipp - integration point number in the mechmat ip array.
  @param im  - index of material type for given ip
  @param ido - index of internal variables for given material in the ipp other array
  
  TKo
*/
void anisodamrot::updateval (long ipp,long im,long ido)
{
  long i;
  long n = Mm->givencompeqother(ipp,im);

  for (i=0;i<n;i++)
    Mm->ip[ipp].eqother[ido+i]=Mm->ip[ipp].other[ido+i];
}
