#include <string.h>
#include "outdiagm.h"
#include "iotools.h"
#include "gtopology.h"
#include "global.h"
#include "probdesc.h"
#include "meshtransfer.h"
#include "mechtop.h"
#include "element.h"
#include "node.h"
#include "intpoints.h"
#include "globmat.h"

/**
  This constructor initializes data to zero values
*/
outdiagm::outdiagm()
{
  pid = NULL; eid = NULL; ipeid = NULL; nif = NULL;
  pu = NULL;
  ipu = NULL;
}


/**
  This destructor deallocates used memory
*/
outdiagm::~outdiagm()
{
  delete [] nif;
  delete [] pid;
  delete [] eid;
  delete [] ipeid;
  delete [] pu;
  delete [] ipu;
}



/** This function reads data from the file given by the pointer to the 
    opened text file.
   
    @param  in - pointer to the opened text file where the data will be read from

    @retval 0 - on success
    @retval 1 - in case read error   
*/
long outdiagm::read(XFILE *in)
{
  long i;
  if (xfscanf (in, "%k%ld", "numunknowns", &npun) != 2)
  {
    fprintf(stderr, "\n\nError reading number of printed unknowns\n");
    fprintf(stderr, " function outdiagm::read, (file %s, line %d)\n", __FILE__,__LINE__);
    return 1;
  }
  xfscanf(in, "%k", "sel_diagstep");
  dstep.read(in);
  if (dstep.st == sel_no)
    return 0;
  nif   = new nodip[npun];
  pid   = new long [npun];
  eid   = new long [npun];
  ipeid = new long [npun];
  pu    = new prunk[npun];
  ipu   = new long [npun];
  memset(nif, 0, sizeof(*nif)*npun);
  memset(pu, 0, sizeof(*pu)*npun);
  memset(ipu, 0, sizeof(*ipu)*npun);
  for (i=0; i<npun; i++)
  {
    pid[i] = eid[i] = ipeid[i] = -1;
    if (xfscanf (in, "%k%m", "point", &nodip_kwdset, (int *)(nif+i)) != 2)
    {
      fprintf(stderr, "\n\nError reading type of point\n");
      fprintf(stderr, " function outdiagm::read, (file %s, line %d)\n", __FILE__,__LINE__);
      return 1;
    }
    switch (nif[i])
    {
      case atnode:
        if (xfscanf (in, "%k%ld", "node", pid+i) != 2)
        {
          fprintf(stderr, "\n\nError reading node id\n");
          fprintf(stderr, " function outdiagm::read, (file %s, line %d)\n", __FILE__,__LINE__);
          return 1;
        }
        pid[i]--;
        break;
      case atip:
        if (xfscanf (in, "%k %ld %k %ld", "elem", eid+i, "ip", ipeid+i) != 4)
        {
          fprintf(stderr, "\n\nError reading element or ip number\n");
          fprintf(stderr, " function outdiagm::read, (file %s, line %d)\n", __FILE__,__LINE__);
          return 1;
        }
        eid[i]--; ipeid[i]--;
        break;
      case atxyz:
        if (xfscanf (in, "%k %le %k %le %k %le", "x", &x, "y", &y, "z", &z) != 6)
        {
          fprintf(stderr, "\n\nError point coordinates\n");
          fprintf(stderr, " function outdiagm::read, (file %s, line %d)\n", __FILE__,__LINE__);
          return 1;
        }
        pid[i] = adjacnode (Gtm,x,y,z);
        break;
      default:
        fprintf(stderr, "\n\nUnknown type of point is required in function outdiagm::read\n");
        fprintf(stderr, " (file %s, line %d)\n", __FILE__, __LINE__); 
        return 1;
    }
  
    if (xfscanf (in, "%k%m", "quant_type", &prunk_kwdset, (int *)(pu+i)) != 2)
    {
      fprintf(stderr, "\n\nError reading type of printed unknown\n");
      fprintf(stderr, " function outdiagm::read, (file %s, line %d)\n", __FILE__,__LINE__);
      return 1;
    }
    if ((pu[i] == pr_stepid) || (pu[i] == pr_appload))
      continue;
    if (xfscanf (in, "%k%ld", "compid", ipu+i) != 2)
    {
      fprintf(stderr, "\n\nError reading type of printed unknown\n");
      fprintf(stderr, " function outdiagm::read, (file %s, line %d)\n", __FILE__,__LINE__);
      return 1;
    }
    ipu[i]--;
    if ((nif[i] == atnode) || (nif[i] == atxyz))
    {
      switch (pu[i])
      {
        case pr_strains :
          Mp->straincomp = 1;
          Mp->strainaver = 1;
          break;
        case pr_stresses :
          Mp->stresscomp = 1;
          Mp->stressaver = 1;
          break;
        case pr_react :
          Mp->reactcomp = 1;
          break;
        case pr_other :
          Mp->othercomp = 1;
          Mp->otheraver = 1;
          break;
        default:
          break;
      }
    }
  }
  return 0;
}



/** This function prints data to the file given by the pointer out.
   
    @param  in - pointer to the opened text file where the data will be read from

    @retval 0 - on success
*/
long outdiagm::print(FILE *out)
{
  long i;
  fprintf(out, "%ld ", npun);
  dstep.print(out);
  if (dstep.st == sel_no)
    return 0;

  for (i=0; i<npun; i++)
  {
    fprintf(out, "%d ", int(nif[i]));
    switch (nif[i])
    {
      case atnode:
        fprintf (out, "%ld ", pid[i]+1);
        break;
      case atip:
        fprintf (out, "%ld %ld ", eid[i]+1, ipeid[i]+1);
        break;
      case atxyz:
        fprintf (out, "%e %e %e ", x, y, z);
        break;
      default:
        fprintf(stderr, "\n\nUnknown type of point is required in function outdiagm::print\n");
        fprintf(stderr, " (file %s, line %d)\n", __FILE__, __LINE__); 
        return 1;
    }
    fprintf (out, "%d", int(pu[i]));

    if ((pu[i] == pr_stepid) || (pu[i] == pr_appload))
    {
      fprintf (out, "\n");
      continue;
    }
    fprintf (out, " %ld\n", ipu[i]+1);
  }

  return 0;
}



/**
  Function prints header for diagram file

  @param out - pointer to opened output text file

  @retval 0 - on success
  @retval 1 - in case of unknown required value type
*/
long outdiagm::print_header(FILE *out)
{
  long i;
  for (i=0; i<npun; i++)
  {
    switch (pu[i])
    {
      case pr_displ :
        fprintf(out, "r_%ld ", ipu[i]);
        break;
      case pr_strains :
        fprintf(out, "eps_%ld ", ipu[i]);
        break;
      case pr_stresses :
        fprintf(out, "sig_%ld ", ipu[i]);
        break;
      case pr_forces :
        fprintf(out, "f_%ld ", ipu[i]);
        break;
      case pr_react :
        fprintf(out, "R_%ld ", ipu[i]);
        break;
      case pr_stepid :
        fprintf(out, "step ");
        break;
      case pr_appload :
        fprintf(out, "lambda ");
        break;
      case pr_other :
        fprintf(out, "other_%ld ", ipu[i]);
        break;
      default:
        fprintf(stderr, "\n\nUnknown type of value for diagram is required\n");
        fprintf(stderr, " in node or ip number %ld, (file %s, line %d)\n", pid[i]+1, __FILE__, __LINE__);
        return 1;
    }
  }
  fprintf(out, "\n");
  return 0;
}



/**
  This function prints one row with required values to the output file out.
  It supposes that values in the integration points and nodes are actualized and valid.

  @param out - pointer to opened output text file
  @param lcid - load case id
  @param lambda - coefficient of applied load
  @param istep - index of current step
  @param fi - pointer to array load vector


  @retval 0 - on success
  @retval 1 - in case of wrong ip identification
  @retval 2 - in case of unknown required value type
*/
long outdiagm::printval(FILE *out, long lcid, double lambda, long istep, double *fi)
{
  long i;

  // check if we should print at given step
  if (dstep.presence_id(istep, lambda, Mp->timecon) == 0)
    return 0;

  for (i=0; i<npun; i++)
  {
    if ((pid[i] < 0) && (nif[i] == atip))
    {
      if ((eid[i] < Mt->ne) && (ipeid[i] >= 0) && (ipeid[i] < Mt->give_tnip(eid[i])))
        pid[i] = Mt->elements[eid[i]].ipp[0][0]+ipeid[i];
      else
      {
        fprintf(stderr, "\n\nInvalid element number or element ip number is required in function outdiagm::print_val\n");
        fprintf(stderr, " element %ld, ip %ld (file %s, line %d)\n", eid[i]+1, ipeid[i]+1, __FILE__, __LINE__); 
        return 1;
      }
    }
    switch (pu[i])
    {
      case pr_displ :
        print_displacements(out, lcid, i);
        break;
      case pr_strains :
        print_strains(out, lcid, i);
        break;
      case pr_stresses :
        print_stresses(out, lcid, i);
        break;
      case pr_forces :
        print_forces(out, i, fi);
        break;
      case pr_react :
        print_reactions(out, lcid, i);
        break;
      case pr_stepid :
        fprintf(out, "%ld ", istep);
        break;
      case pr_appload :
	if(Mp->tpr == seconds)
	  fprintf(out, "% e ", lambda);
	if(Mp->tpr == minutes)
	  fprintf(out, "% e ", lambda/60.0);
	if(Mp->tpr == hours)
	  fprintf(out, "% e ", lambda/3600.0);
	if(Mp->tpr == days)
	  fprintf(out, "% e ", lambda/86400.0);
        break;
      case pr_other :
        print_others(out, lcid, i);
        break;
      default:
        fprintf(stderr, "\n\nUnknown type of value for diagram is required\n");
        fprintf(stderr, " in node or ip number %ld, (file %s, line %d)\n", pid[i]+1, __FILE__, __LINE__);
        return 2;
    }
  }
  fprintf(out, "\n");

  return 0;
}



/**
  This function prints one row with required values to the output file out without respect to
  to selected step/time (forced printing). It supposes that values in the integration points 
  and nodes are actualized and valid.

  @param out - pointer to opened output text file
  @param lcid - load case id
  @param lambda - coefficient of applied load
  @param istep - index of current step
  @param fi - pointer to array load vector


  @retval 0 - on success
  @retval 1 - in case of wrong ip identification
  @retval 2 - in case of unknown required value type
*/
long outdiagm::printval_forced(FILE *out, long lcid, double lambda, long istep, double *fi)
{
  long i;


  for (i=0; i<npun; i++)
  {
    if ((pid[i] < 0) && (nif[i] == atip))
    {
      if ((eid[i] < Mt->ne) && (ipeid[i] >= 0) && (ipeid[i] < Mt->give_tnip(eid[i])))
        pid[i] = Mt->elements[eid[i]].ipp[0][0]+ipeid[i];
      else
      {
        fprintf(stderr, "\n\nInvalid element number or element ip number is required in function outdiagm::print_val\n");
        fprintf(stderr, " element %ld, ip %ld (file %s, line %d)\n", eid[i]+1, ipeid[i]+1, __FILE__, __LINE__); 
        return 1;
      }
    }
    switch (pu[i])
    {
      case pr_displ :
        print_displacements(out, lcid, i);
        break;
      case pr_strains :
        print_strains(out, lcid, i);
        break;
      case pr_stresses :
        print_stresses(out, lcid, i);
        break;
      case pr_forces :
        print_forces(out, i, fi);
        break;
      case pr_react :
        print_reactions(out, lcid, i);
        break;
      case pr_stepid :
        fprintf(out, "%ld ", istep);
        break;
      case pr_appload :
	if(Mp->tpr == seconds)
	  fprintf(out, "% e ", lambda);
	if(Mp->tpr == minutes)
	  fprintf(out, "% e ", lambda/60.0);
	if(Mp->tpr == hours)
	  fprintf(out, "% e ", lambda/3600.0);
	if(Mp->tpr == days)
	  fprintf(out, "% e ", lambda/86400.0);
        break;
      case pr_other :
        print_others(out, lcid, i);
        break;
      default:
        fprintf(stderr, "\n\nUnknown type of value for diagram is required\n");
        fprintf(stderr, " in node or ip number %ld, (file %s, line %d)\n", pid[i]+1, __FILE__, __LINE__);
        return 2;
    }
  }
  fprintf(out, "\n");

  return 0;
}



/**
  This function prints displacement at required point to the output file out.
  It supposes that values in the nodes are actualized and valid.

  @param out - pointer to opened output text file
  @param lcid - load case id
  @param idp - required component of displacement

  @retval 0 - on success
  @retval 1 - in case that value at integration point is required
*/
long outdiagm::print_displacements(FILE *out, long lcid, long idp)
{
  double *r;
  switch (nif[idp])
  {
    case atxyz:
    case atnode:
      r = new double [Gtm->gnodes[pid[idp]].ndofn];
      noddispl (lcid, r, pid[idp]);
      fprintf(out, "% .15e ", r[ipu[idp]]);
      delete [] r;
      break;
    case atip:
      fprintf (stderr, "\n\nUnsupported combination of printed values is required -\n"); 
      fprintf(stderr, " displacements in ip number %ld, outdiagm::print_displacements (file %s, line %d)\n", pid[idp]+1, __FILE__, __LINE__);
      return 1;
    default:
      fprintf(stderr, "\n\nUnknown type of point is required in function outdiagm::print_displacements\n");
      fprintf(stderr, " point number %ld (file %s, line %d)\n", pid[idp]+1, __FILE__, __LINE__); 
      return 1;
  }
  return  0;
}


/**
  This function prints strain at required point to the output file out.
  It supposes that values in the integration points and nodes are actualized and valid.

  @param out - pointer to opened output text file
  @param lcid - load case id
  @param idp - required component of strain

  @retval 0 - on success
  @retval 1 - in case that unknown type of point is required
  @retval 2 - in case that unknown component of strain is required
*/
long outdiagm::print_strains(FILE *out, long lcid, long idp)
{
  switch (nif[idp])
  {
    case atxyz:
    case atnode:
      if (ipu[idp] < Mt->nodes[pid[idp]].ncompstr)
        fprintf(out, "% .15e ", Mt->nodes[pid[idp]].strain[ipu[idp]]);
      else
      {
        fprintf(stderr, "\n\nInvalid component number is required in function outdiagm::print_strains\n");
        fprintf(stderr, " node number %ld (file %s, line %d)\n", pid[idp]+1, __FILE__, __LINE__); 
        return 2;
      }
      break;
    case atip:
      if (ipu[idp] < Mm->ip[pid[idp]].ncompstr)
        fprintf(out, "% .15e ", Mm->ip[pid[idp]].strain[ipu[idp]]);
      else
      {
        fprintf(stderr, "\n\nInvalid component number is required in function outdiagm::print_strains\n");
        fprintf(stderr, " integration point number %ld (file %s, line %d)\n", pid[idp]+1, __FILE__, __LINE__); 
        return 2;
      }
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of point is required in function outdiagm::print_strains\n");
      fprintf(stderr, " point number %ld (file %s, line %d)\n", pid[idp]+1, __FILE__, __LINE__); 
      return 1;
  }
  return 0;
}


/**
  This function prints stress at required point to the output file out.
  It supposes that values in the integration points and nodes are actualized and valid.

  @param out - pointer to opened output text file
  @param lcid - load case id
  @param idp - required component of stress

  @retval 0 - on success
  @retval 1 - in case that unknown type of point is required
  @retval 2 - in case that unknown component of strain is required
*/
long outdiagm::print_stresses(FILE *out, long lcid, long idp)
{
  switch (nif[idp])
  {
    case atxyz:
    case atnode:
      if (ipu[idp] < Mt->nodes[pid[idp]].ncompstr)
        fprintf(out, "% .15e ", Mt->nodes[pid[idp]].stress[ipu[idp]]);
      else
      {
        fprintf(stderr, "\n\nInvalid component number is required in function outdiagm::print_stresses\n");
        fprintf(stderr, " node number %ld (file %s, line %d)\n", pid[idp]+1, __FILE__, __LINE__); 
        return 2;
      }
      break;
    case atip:
      if (ipu[idp] < Mm->ip[pid[idp]].ncompstr)
        fprintf(out, "% .15e ", Mm->ip[pid[idp]].stress[ipu[idp]]);
      else
      {
        fprintf(stderr, "\n\nInvalid component number is required in function outdiagm::print_stresses\n");
        fprintf(stderr, " integration point number %ld (file %s, line %d)\n", pid[idp]+1, __FILE__, __LINE__); 
        return 2;
      }
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of point is required in function outdiagm::print_stresses\n");
      fprintf(stderr, " point number %ld (file %s, line %d)\n", pid[idp]+1, __FILE__, __LINE__); 
      return 1;
  }
  return 0;
}


/**
  This function prints forces at required node to the output file out.
  Value from load vector fi is used and printed.

  @param out - pointer to opened output text file
  @param idp - required component of force

  @retval 0 - on success
  @retval 1 - in case that point type is atip
  @retval 2 - in case that unknown component of force is required
*/
long outdiagm::print_forces(FILE *out, long idp, double *fi)
{
  long ii;
  switch (nif[idp])
  {
    case atxyz:
    case atnode:
      ii=Mt->give_dof(pid[idp], ipu[idp]);
      if (ii > 0)
        fprintf(out, "% .15e ", fi[ii-1]);
      else      
      {
        fprintf(stderr, "\n\nInvalid component number is required in function outdiagm::print_forces\n");
        fprintf(stderr, " node number %ld (file %s, line %d)\n", pid[idp]+1, __FILE__, __LINE__); 
        return 2;
      }
      break;
    case atip:
      fprintf (stderr, "\n\nUnsupported combination of printed values is required -\n"); 
      fprintf(stderr, " forces in ip number %ld, outdiagm::print_forces (file %s, line %d)\n", pid[idp]+1, __FILE__, __LINE__);
      return 1;
    default:
      fprintf(stderr, "\n\nUnknown type of point is required in function outdiagm::print_forces\n");
      fprintf(stderr, " point number %ld (file %s, line %d)\n", pid[idp]+1, __FILE__, __LINE__); 
      return 1;
  }
  return 0;
}


/**
  This function prints reaction at required node to the output file out.
  It supposes that values in the nodes are actualized and valid.

  @param out - pointer to opened output text file
  @param lcid - load case id
  @param idp - required component of reaction

  @retval 0 - on success
  @retval 1 - in case that unknown type of point is required
*/
long outdiagm::print_reactions(FILE *out, long lcid, long idp)
{
  switch (nif[idp])
  {
    case atxyz:
    case atnode:
      if (Mt->nodes[pid[idp]].react)
        fprintf(out, "% .15e ", Mt->nodes[pid[idp]].r[ipu[idp]]);
      break;
    case atip:
      fprintf (stderr, "\n\nUnsupported combination of printed values is required -\n"); 
      fprintf(stderr, " forces in ip number %ld, outdiagm::print_reactions (file %s, line %d)\n", pid[idp]+1, __FILE__, __LINE__);
      return 1;
    default:
      fprintf(stderr, "\n\nUnknown type of point is required in function outdiagm::print_reactions\n");
      fprintf(stderr, " point number %ld (file %s, line %d)\n", pid[idp]+1, __FILE__, __LINE__); 
      return 1;
  }
  return 0;
}


/**
  This function prints other value at required point to the output file out.
  It supposes that values in the integration points and nodes are actualized and valid.

  @param out - pointer to opened output text file
  @param lcid - load case id
  @param idp - required component of other array

  @retval 0 - on success
  @retval 1 - in case that unknown type of point is required
  @retval 2 - in case that unknown component of strain is required
*/
long outdiagm::print_others(FILE *out, long lcid, long idp)
{
  switch (nif[idp])
  {
    case atxyz:
    case atnode:
      if (ipu[idp] < Mt->nodes[pid[idp]].ncompother)
        fprintf(out, "% .15e ", Mt->nodes[pid[idp]].other[ipu[idp]]);
      else
      {
        fprintf(stderr, "\n\nInvalid component number is required in function outdiagm::print_others\n");
        fprintf(stderr, " node number %ld (file %s, line %d)\n", pid[idp]+1, __FILE__, __LINE__); 
        return 2;
      }
      break;
    case atip:
      if (ipu[idp] < Mm->ip[pid[idp]].ncompeqother)
        fprintf(out, "% .15e ", Mm->ip[pid[idp]].eqother[ipu[idp]]);
      else
      {
        fprintf(stderr, "\n\nInvalid component number is required in function outdiagm::print_others\n");
        fprintf(stderr, " integration point number %ld (file %s, line %d)\n", pid[idp]+1, __FILE__, __LINE__); 
        return 2;
      }
      break;
    default:
      fprintf(stderr, "\n\nUnknown type of point is required in function outdiagm::print_others\n");
      fprintf(stderr, " point number %ld (file %s, line %d)\n", pid[idp]+1, __FILE__, __LINE__); 
      return 1;
  }
  return 0;
}
